{*****************************************************************************}
{                                                                             }
{ SR Voices Extractor - converter sound and music from Soul Reaver PC-version }
{                                                                             }
{ Written by Andrey Grachev (http://wheelofdoom.ucoz.com/)                     }
{                                                                             }
{ This program is free software; you can redistribute it and/or               }
{ modify it under the terms of the GNU General Public License                 }
{ as published by the Free Software Foundation; either version 2              }
{ of the License, or (at your option) any later version.                      }
{                                                                             }
{ This program is distributed in the hope that it will be useful,             }
{ but WITHOUT ANY WARRANTY; without even the implied warranty of              }
{ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                        }
{                                                                             }
{ See the GNU General Public License for more details.                        }
{                                                                             }
{ You should have received a copy of the GNU General Public License           }
{ along with this program; if not, write to the Free Software                 }
{ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. }
{                                                                             }
{*****************************************************************************}

program voext;

{$APPTYPE CONSOLE}
{$R Resource/voextres.res}

uses
  Windows, SysUtils, Classes, StrUtils;

type
  {voices.dat / music.dat structure: Offset DWORD, File size DWORD}
  TSoundInfoSR = packed record
    FNumber: Cardinal;
    Offset: array of Cardinal;
    FileSize: array of Cardinal;
  end;

  {WAVE header}
  TWaveHeader = record
    RIFFHeader: array[0..3] of Char;  {"RIFF"}
    FileSize: Integer;                {File size - 8}
    WAVEHeader: array[0..3] of Char;  {"WAVE"}
    fmtHeader: array[0..3] of Char;   {"fmt "}
    fmtSize: Integer;                 {16}
    fmtId: Word;                      {For PCM - 1}
    Channels: Word;                   {1 - mono; 2 - stereo}
    SampleRate: Integer;              {11025, 22050, 44100 Hz}
    BytesPerSecond: Integer;          {(Channels*SampleRate*BitsPerSample)/8}
    BlockAlign: Word;                 {(Channels*BitsPerSample)/8}
    BitsPerSample: Word;              {8, 16, 32}
    DataHeader: array[0..3] of Char;  {"data"}
    DataSize: Integer;                {Data size in bytes}
  end;

const
  SR1_VOICE_SIZE_PC = $90FC696;      {152028822}
  SR1_MUSIC_SIZE_PC = $6AA1514;      {111809812}

  RIFFHdr = 'RIFF';
  WAVEHdr = 'WAVE';
  fmtHdr = 'fmt ';
  DATA_CHUNK = 'data';

  VOICE_DIR = 'Voices';
  MUSIC_DIR = 'Music';

var
  FileSound: TSoundInfoSR;
  FName, WaveName: string;
  Extension: string = '.wav';
  DirName, ResName: string;
  Prefix: string;
  i: Word;
  FNum: Word = 0;
  Persent: Word = 0;
  SourceFile: TFileStream;
  DestinationFile: TFileStream;
  Directory: string;

//Adding a backslash to the directory name
function AddBackSlash(Path: string): string;
begin
  if Path[Length(Path)] <> '\' then
    Result := Path + '\'
  else
    Result := Path;
end;

//Output indicating progress on the screen
function ShowProgress(Persent: Integer): string;
var
  Backspace: string;
begin
  Backspace := DupeString(#8, Length(IntToStr(Persent)) + 1);
  Write(Persent, '%', Backspace);
end;

//Building header of WAVE
function BuildHeader(var WaveHeader: TWaveHeader; DataSize: Integer;
  Channels: Byte): Boolean;
begin
  try
    Result := True;
    WaveHeader.RIFFHeader := RIFFHdr;
    WaveHeader.FileSize := DataSize + SizeOf(WaveHeader) - 8;
    WaveHeader.WAVEHeader := WAVEHdr;
    WaveHeader.fmtHeader := fmtHdr;
    WaveHeader.fmtSize := 16;
    WaveHeader.fmtId := 1;
    WaveHeader.Channels := Channels;
    WaveHeader.SampleRate := 22050;
    WaveHeader.BitsPerSample := 16;
    WaveHeader.BytesPerSecond := (WaveHeader.Channels *
      WaveHeader.SampleRate * WaveHeader.BitsPerSample) div 8;
    WaveHeader.BlockAlign := (WaveHeader.Channels *
      WaveHeader.BitsPerSample) div 8;
    WaveHeader.DataHeader := DATA_CHUNK;
    WaveHeader.DataSize := DataSize;
  except
    Result := False;
    Exit;
  end;
end;

//Reading the file allocation table in voices.dat from resource
function ReadSoundSR(var FileData: TSoundInfoSR; ResName: string): Boolean;
var
  Stream: TStream;
  HResInfo, MemHandle: THandle;
  ResSize, FNumber: Longint;
  ResPtr: PByte;
  i: Cardinal;
begin
  try
    Result := True;
    HResInfo := FindResource(HInstance, PChar(ResName), 'BIN');
    ResSize := SizeOfResource(HInstance, HResInfo);
    MemHandle := LoadResource(HInstance, HResInfo);
    ResPtr := LockResource(MemHandle);
    Stream := TMemoryStream.Create;
    try
      Stream.Size := ResSize;
      Stream.Write(ResPtr^, ResSize);
      Stream.Position := 0;
      FNumber := Stream.Size div 8;
      FileData.FNumber := FNumber;
      if FNumber > 0 then
      begin
        SetLength(FileData.Offset, FNumber);
        SetLength(FileData.FileSize, FNumber);
        for i := 0 to FNumber - 1 do
        begin
          Stream.Read(FileData.Offset[i], 4);
          Stream.Read(FileData.FileSize[i], 4);
        end;
      end;
    finally
      Stream.Free;
      FreeResource(MemHandle);
    end;
  except
    Result := False;
    Exit;
  end;
end;

//Extracting a file from voices.dat and convert it into WAVE-file
function ExtractFile(var SourceFile, DestinationFile: TFileStream;
  Offset: Cardinal; Size: Cardinal): Boolean;
var
  WaveHeader: TWaveHeader;
  Channels: Byte;
begin
  case SourceFile.Size of
    SR1_VOICE_SIZE_PC: Channels := 1;
    SR1_MUSIC_SIZE_PC: Channels := 2;
  else
    Channels := 1;
  end;

  try
    Result := True;
    SourceFile.Seek(Offset, soFromBeginning);
    if Size > 0 then
      if BuildHeader(WaveHeader, Size, Channels) then
      begin
        DestinationFile.Write(WaveHeader, SizeOf(WaveHeader));
        DestinationFile.CopyFrom(SourceFile, Size);
      end;
  except
    Result := False;
    Exit;
  end;
end;

{-----------------------------------------------------------------}

begin
  Writeln('Converter sound and music from Soul Reaver PC-version');
  Writeln('-= Written by Andrey Grachev =-', #13#10);

  if ParamCount <> 1 then
  begin
    Writeln('Usage voext voices.dat');
    Exit;
  end;

  if not FileExists(ParamStr(1)) then
  begin
    Writeln('File ', ParamStr(1), ' not found');
    Exit;
  end;

  SourceFile := TFileStream.Create(ParamStr(1), fmShareDenyNone);

  if (SourceFile.Size <> SR1_VOICE_SIZE_PC) and
     (SourceFile.Size <> SR1_MUSIC_SIZE_PC) then
  begin
    Writeln('Wrong file size');
    Exit;
  end;

  Writeln('Open ', ParamStr(1));
  Writeln('File size: ', SourceFile.Size, ' bytes');

  case SourceFile.Size of
    SR1_VOICE_SIZE_PC:
      begin
        DirName := VOICE_DIR;
        ResName := 'VOICESPC';
        Prefix := 'voice';
      end;
    SR1_MUSIC_SIZE_PC:
      begin
        DirName := MUSIC_DIR;
        ResName := 'MUSIC';
        Prefix := 'music';
      end;
  end;

  if not DirectoryExists(DirName) then
    if not CreateDir(DirName) then
    begin
      WriteLn('Cannot create output directory');
      Exit;
    end;

  if ReadSoundSR(FileSound, ResName) then
  begin
    Write('Extracting PCM-files... ');
    GetDir(0, Directory);
    for i := 0 to FileSound.FNumber - 1 do
    begin
      WaveName := Prefix + DupeString('0', 3 - Length(IntToStr(i + 1))) +
        IntToStr(i + 1) + Extension;
      FName := AddBackSlash(Directory) + AddBackSlash(DirName) + WaveName;
      DestinationFile := TFileStream.Create(FName, fmCreate);
      try
        if not ExtractFile(SourceFile, DestinationFile, FileSound.Offset[i],
          FileSound.FileSize[i]) then
            Writeln('Cannot extract ' + FName)
        else
        begin
          Inc(FNum);
          Persent := (100 * FNum) div FileSound.FNumber;
          ShowProgress(Persent);
        end;
      finally
        DestinationFile.Free;
      end;
    end;
  end;

  SourceFile.Free;
  Writeln(#13#10, 'Total files: ', FNum);
  Writeln('All is done!', #13#10);
end.
